import dayjs from 'dayjs';
import { isDate, isISO8601, isNumeric, isInt, isURL } from 'validator';

class Exception extends Error {
    status;
    context;
    constructor(status, message, 
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    context) {
        super(message);
        this.status = status;
        this.context = context;
    }
}

class ValidationException extends Exception {
    errors;
    constructor(errors) {
        super(422, 'バリデーションエラーが発生しました', errors);
        this.errors = errors;
    }
}

/* eslint-disable @typescript-eslint/no-non-null-assertion,@typescript-eslint/no-explicit-any */
class ValueObject {
    _input;
    _setInner = false;
    _inner;
    _setOutput = false;
    _output;
    constructor(_input) {
        this._input = _input;
    }
    fromInput() {
        return this.input;
    }
    toOutput() {
        return this.inner;
    }
    get input() {
        return this._input;
    }
    get inner() {
        if (!this._setInner) {
            this._setInner = true;
            this._inner = this.fromInput();
        }
        // biome-ignore lint/style/noNonNullAssertion:
        return this._inner;
    }
    get value() {
        if (!this._setOutput) {
            this._setOutput = true;
            this._output = this.toOutput();
            Object.freeze(this._output);
        }
        // biome-ignore lint/style/noNonNullAssertion:
        return this._output;
    }
    equals(value) {
        return this.compare(value) === 0;
    }
    validate(name, prev) {
        const errors = this.getErrors(name, prev);
        if (errors?.length) {
            throw new ValidationException(errors.reduce((acc, error) => {
                acc[error.name] = [
                    ...new Set([...(acc[error.name] ?? []), error.error]),
                ];
                return acc;
            }, {}));
        }
    }
}

var index$1 = /*#__PURE__*/Object.freeze({
    __proto__: null,
    default: ValueObject
});

// Inner type = string: consider serialization
class DateObject extends ValueObject {
    fromInput() {
        return dayjs(this.input).toISOString();
    }
    toOutput() {
        return dayjs(this.inner);
    }
    getErrors(name) {
        if (typeof this.input === 'string' &&
            !(isDate(this.input) || isISO8601(this.input))) {
            return [{ name, error: '日付の形式が正しくありません' }];
        }
        return undefined;
    }
    compare(value) {
        if (this.value.isSame(value.value)) {
            return 0;
        }
        if (this.value.isBefore(value.value)) {
            return -1;
        }
        return 1;
    }
}

var date = /*#__PURE__*/Object.freeze({
    __proto__: null,
    default: DateObject
});

class Flags extends ValueObject {
    getErrors(name) {
        if (!this.flagTypes.includes(this.input)) {
            return [{ name, error: `定義されていないフラグです: ${this.input}` }];
        }
        return undefined;
    }
    compare(value) {
        return this.value.localeCompare(value.value);
    }
}

var flags = /*#__PURE__*/Object.freeze({
    __proto__: null,
    default: Flags
});

class Float extends ValueObject {
    fromInput() {
        if (typeof this.input === 'number') {
            return this.input;
        }
        let num = Number.parseFloat(this.input);
        const max = this.getMaxNumber();
        if (max !== undefined && num > max && this.isTruncateMode()) {
            num = Math.min(max, num);
        }
        const min = this.getMinNumber();
        if (min !== undefined && num < min && this.isTruncateMode()) {
            num = Math.max(min, num);
        }
        return num;
    }
    getMaxNumber() {
        return undefined;
    }
    getMinNumber() {
        return undefined;
    }
    isTruncateMode() {
        return false;
    }
    getErrors(name) {
        if (typeof this.input === 'string' && !isNumeric(this.input)) {
            return [{ name, error: '数値の形式が正しくありません' }];
        }
        const num = this.fromInput();
        const max = this.getMaxNumber();
        if (max !== undefined && num > max) {
            return [{ name, error: `${max}以下の値を入力してください` }];
        }
        const min = this.getMinNumber();
        if (min !== undefined && num < min) {
            return [{ name, error: `${min}以上の値を入力してください` }];
        }
        return undefined;
    }
    compare(value) {
        const diff = this.value - value.value;
        if (diff < 0)
            return -1;
        if (diff > 0)
            return 1;
        return 0;
    }
}

var float = /*#__PURE__*/Object.freeze({
    __proto__: null,
    default: Float
});

class Int extends Float {
    fromInput() {
        return Math.floor(super.fromInput());
    }
    getErrors(name) {
        const results = super.getErrors(name);
        if (results?.length) {
            return results;
        }
        if (typeof this.input === 'string' && !isInt(this.input)) {
            return [{ name, error: '整数の形式が正しくありません' }];
        }
        const num = this.fromInput();
        if (!Number.isSafeInteger(num)) {
            return [{ name, error: '有効な整数ではありません' }];
        }
        return undefined;
    }
}

var int = /*#__PURE__*/Object.freeze({
    __proto__: null,
    default: Int
});

class InvalidValueException extends Exception {
    constructor(target, reason) {
        super(400, '無効な値です', { target, reason });
    }
}

class IntId extends ValueObject {
    fromInput() {
        const _input = this.input;
        if (_input === null) {
            return null;
        }
        if (typeof _input === 'number') {
            return Math.floor(_input);
        }
        return Math.floor(Number.parseFloat(_input));
    }
    toOutput() {
        if (this.inner === null) {
            throw new InvalidValueException('id');
        }
        return super.toOutput();
    }
    isSet() {
        return this.input !== null && this.input !== undefined;
    }
    getErrors(name) {
        const num = this.fromInput();
        if (num === null) {
            return undefined;
        }
        if (typeof this.input === 'string' && !isInt(this.input)) {
            return [{ name, error: '整数の形式が正しくありません' }];
        }
        if (!Number.isSafeInteger(num)) {
            return [{ name, error: '有効な整数ではありません' }];
        }
        return undefined;
    }
    compare(value) {
        const diff = this.value - value.value;
        if (diff < 0)
            return -1;
        if (diff > 0)
            return 1;
        return 0;
    }
}

var intId = /*#__PURE__*/Object.freeze({
    __proto__: null,
    default: IntId
});

class StringId extends ValueObject {
    fromInput() {
        if (this.input === null) {
            return null;
        }
        return `${this.input}`;
    }
    toOutput() {
        if (this.inner === null) {
            throw new InvalidValueException('id');
        }
        return super.toOutput();
    }
    isSet() {
        return this.input !== null && this.input !== undefined;
    }
    getErrors(name) {
        const text = this.fromInput();
        if (text === null) {
            return undefined;
        }
        if (!text.length) {
            return [{ name, error: '値を指定してください' }];
        }
        return undefined;
    }
    compare(value) {
        return this.value.localeCompare(value.value);
    }
}

var stringId = /*#__PURE__*/Object.freeze({
    __proto__: null,
    default: StringId
});

class Text extends ValueObject {
    fromInput() {
        if (typeof this.input === 'number') {
            return `${this.input}`;
        }
        if (this.input) {
            return `${this.input}`;
        }
        return '';
    }
    getValidationMinLength() {
        return undefined;
    }
    getValidationMaxLength() {
        return undefined;
    }
    getErrors(name) {
        const text = this.fromInput();
        const results = [];
        if (!text.length) {
            results.push({ name, error: '値を指定してください' });
        }
        else {
            const min = this.getValidationMinLength();
            if (min && text.length < min) {
                results.push({ name, error: `${min}文字より長く入力してください` });
            }
        }
        const max = this.getValidationMaxLength();
        if (max && text.length > max) {
            results.push({ name, error: `${max}文字より短く入力してください` });
        }
        return results;
    }
    compare(value) {
        return this.value.localeCompare(value.value);
    }
}

var text = /*#__PURE__*/Object.freeze({
    __proto__: null,
    default: Text
});

class Url extends ValueObject {
    getErrors(name) {
        if (!isURL(this.input)) {
            return [{ name, error: 'URLの形式が正しくありません' }];
        }
        return undefined;
    }
    compare(value) {
        return this.value.localeCompare(value.value);
    }
}

var url = /*#__PURE__*/Object.freeze({
    __proto__: null,
    default: Url
});

class InvalidUsage extends Exception {
    constructor(reason) {
        super(500, '使用方法が正しくありません', { reason });
    }
}

class Collection {
    collections;
    static _isCreating = false;
    // create メソッドの this コンテキストのせいで protected にはできない
    /**
     * @deprecated create 経由で生成
     */
    constructor(collections) {
        this.collections = collections;
        if (!Collection._isCreating) {
            throw new InvalidUsage('create経由で生成してください');
        }
        Object.freeze(this.collections);
    }
    [Symbol.iterator]() {
        return this.collections.values();
    }
    find(filter) {
        return this.collections.find(filter);
    }
    filter(filter) {
        return this.collections.filter(filter);
    }
    isEmpty() {
        return !this.collections.length;
    }
    getErrors(prev) {
        const getPrev = (item) => (prev?.collections ?? []).find((x) => item.equals(x));
        const errors = this.collections.reduce((acc, item, index) => ({
            // biome-ignore lint/performance/noAccumulatingSpread:
            ...acc,
            ...Object.fromEntries(Object.entries(item.getErrors(getPrev(item))).map(([key, value]) => [
                `${key}[${index}]`,
                value,
            ])),
        }), {});
        if (Object.keys(errors).length) {
            return errors;
        }
        return undefined;
    }
    validate(prev) {
        const errors = this.getErrors(prev);
        if (errors) {
            throw new ValidationException(errors);
        }
    }
    static create(collections) {
        try {
            Collection._isCreating = true;
            // biome-ignore lint/complexity/noThisInStatic:
            return new this(collections);
        }
        finally {
            Collection._isCreating = false;
        }
    }
}

var collection = /*#__PURE__*/Object.freeze({
    __proto__: null,
    default: Collection
});

/* eslint-disable @typescript-eslint/no-explicit-any */
class Entity {
    static _isCreating = false;
    constructor() {
        if (!Entity._isCreating) {
            throw new InvalidUsage();
        }
    }
    static _create(...args) {
        try {
            Entity._isCreating = true;
            // biome-ignore lint/complexity/noThisInStatic:
            const instance = new this(...args);
            instance.validate();
            return instance;
        }
        finally {
            Entity._isCreating = false;
        }
    }
    static _reconstruct(...args) {
        try {
            Entity._isCreating = true;
            // biome-ignore lint/complexity/noThisInStatic:
            return new this(...args);
        }
        finally {
            Entity._isCreating = false;
        }
    }
    static _update(target, ...args) {
        try {
            Entity._isCreating = true;
            // biome-ignore lint/complexity/noThisInStatic:
            const instance = new this(...args);
            instance.validate(target);
            return instance;
        }
        finally {
            Entity._isCreating = false;
        }
    }
    getErrors(prev) {
        return Object.keys(this).reduce((acc, key) => {
            const member = this[key];
            const prevValue = prev
                ? (prev[key] ?? undefined)
                : undefined;
            if (member && member instanceof Collection) {
                const errors = member.getErrors(prevValue);
                if (errors) {
                    return Object.entries(errors).reduce((acc, [key, value]) => {
                        acc[key] = [...new Set([...(acc[key] ?? []), ...value])];
                        return acc;
                    }, acc);
                }
            }
            if (member && member instanceof ValueObject) {
                const name = key.replace(/^_/, '');
                const errors = member.getErrors(name, prevValue);
                if (errors?.length) {
                    return errors.reduce((acc, error) => {
                        acc[error.name] = [
                            ...new Set([...(acc[error.name] ?? []), error.error]),
                        ];
                        return acc;
                    }, acc);
                }
            }
            return acc;
        }, {});
    }
    validate(prev) {
        const errors = this.getErrors(prev);
        if (Object.keys(errors).length) {
            throw new ValidationException(errors);
        }
    }
}

var index = /*#__PURE__*/Object.freeze({
    __proto__: null,
    default: Entity
});

export { collection as Collection, date as DateObject, index as Entity, flags as Flags, float as Float, int as Int, intId as IntId, stringId as StringId, text as Text, url as Url, index$1 as ValueObject };
